@@ -31,33 +31,57 @@ class DoesNotExists(OneDriveInteractionError):
 
 class OneDriveHTTPClient(object):
 
-	def request( self, url, method='get', data=None,
-				files=None, raw=False, headers=dict(), raise_for=dict(),
-				session=None ):
+	#: Extra keywords to pass to each "requests.Session.request()" call.
+	#: For full list of these see:
+	#:  http://docs.python-requests.org/en/latest/api/#requests.Session.request
+	request_extra_keywords = None # Example: dict(timeout=(20.0, 10.0))
+
+	#: Keywords to pass to "requests.adapters.HTTPAdapter" subclass init.
+	#: Only used with later versions of "requests" than 1.0.0 (where adapters were introduced).
+	request_adapter_settings = None # Example: dict(max_retries=2)
+
+	_requests_setup_done = False
+
+	def _requests_setup(self, requests, **adapter_kws):
+		session, requests_version = None, requests.__version__
+		log.debug('Using "requests" module version: %r', requests_version)
+		try: requests_version = tuple(it.imap(int, requests_version.split('.')))
+		except: requests_version = 999, 0, 0 # most likely some future version
+
+		if requests_version >= (1, 0, 0):
+			session = requests.Session()
+			session.mount('https://', requests.adapters.HTTPAdapter(**adapter_kws))
+
+		elif requests_version < (0, 14, 0):
+			raise RuntimeError( (
+				'Version of the "requests" python module (used by python-onedrive)'
+					' is incompatible - need at least 0.14.0, but detected {}.'
+					' Please update it (or file an issue if it worked before).' )\
+				.format(requests.__version__) )
+
+		self._requests_setup_done = True
+		return session
+
+	def request( self, url, method='get', data=None, files=None,
+				raw=False, headers=dict(), raise_for=dict(), session=None ):
 		'''Make synchronous HTTP request.
 			Can be overidden to use different http module (e.g. urllib2, twisted, etc).'''
 		import requests # import here to avoid dependency on the module
 
-		requests_version = getattr(requests, '__version__', None)
-		if requests_version:
-			log.debug('Using "requests" module version: %r', requests_version)
-			try: requests_version = tuple(it.imap(int, requests_version.split('.')))
-			except: requests_version = 999, 0, 0 # most likely some future version
-			if requests_version < (0, 14, 0):
-				raise RuntimeError( (
-					'Version of the "requests" python module (used by python-onedrive)'
-						' is incompatible - need at least 0.14.0, but detected {}.'
-						' Please update it (or file an issue if it worked before).' )\
-					.format(requests.__version__) )
+		if not self._requests_setup_done:
+			# (hopefully) temporary fix for https://github.com/mk-fg/python-onedrive/issues/1
+			patched_session = self._requests_setup(
+				requests, **(self.request_adapter_settings or dict()) )
+			if patched_session is not None: self._requests_session = patched_session
 
 		if session is None:
-			try: session = self._requests_session
-			except AttributeError: session = self._requests_session = requests.session()
+			session = getattr(self, '_requests_session', None)
+			if not session: session = self._requests_session = requests.session()
 		elif not session: session = requests
 
 		method = method.lower()
-		kwz, func = dict(), getattr(
-			session, method, ft.partial(session.request, method.upper()) )
+		kwz, func = dict(), ft.partial(
+			session.request, method.upper(), **(self.request_extra_keywords or dict()) )
 		if data is not None:
 			if method in ['post', 'put']: kwz['data'] = data
 			else:
@@ -187,6 +211,11 @@ class OneDriveAPIWrapper(OneDriveAuth):
 
 	api_url_base = 'https://apis.live.net/v5.0/'
 
+	#: Limit on file uploads via single PUT request, imposed by the API.
+	#: Used to opportunistically fallback to BITS API
+	#:  (uploads via several http requests) in the "put" method.
+	api_put_max_bytes = int(95e6)
+
 	def _api_url( self, path, query=dict(),
 			pass_access_token=True, pass_empty_values=False ):
 		query = query.copy()
@@ -234,18 +263,25 @@ class OneDriveAPIWrapper(OneDriveAuth):
 			return self.request(api_url(), **request_kwz)
 
 	def get_quota(self):
-		'Get OneDrive object, representing quota.'
+		'Get OneDrive object representing quota.'
 		return self('me/skydrive/quota')
 
+	def get_user_data(self):
+		'Get OneDrive object representing user metadata (including user "id").'
+		return self('me')
+
+	def get_user_id(self):
+		'Returns "id" of a OneDrive user.'
+		return self.get_user_data()['id']
+
 	def listdir(self, folder_id='me/skydrive', limit=None, offset=None):
-		'Get OneDrive object, representing list of objects in a folder.'
+		'Get OneDrive object representing list of objects in a folder.'
 		return self(ujoin(folder_id, 'files'), dict(limit=limit, offset=offset))
 
 	def info(self, obj_id='me/skydrive'):
 		'''Return metadata of a specified object.
 			See http://msdn.microsoft.com/en-us/library/live/hh243648.aspx
-				for the list and description of metadata keys for
-				each object type.'''
+				for the list and description of metadata keys for each object type.'''
 		return self(obj_id)
 
 	def get(self, obj_id, byte_range=None):
@@ -258,7 +294,8 @@ class OneDriveAPIWrapper(OneDriveAuth):
 		return self(ujoin(obj_id, 'content'), dict(download='true'),
 					raw=True, **kwz)
 
-	def put(self, path_or_tuple, folder_id='me/skydrive', overwrite=None, downsize=None):
+	def put( self, path_or_tuple, folder_id='me/skydrive',
+			overwrite=None, downsize=None, bits_api_fallback=True ):
 		'''Upload a file (object), possibly overwriting (default behavior)
 				a file with the same "name" attribute, if it exists.
 
@@ -270,8 +307,13 @@ class OneDriveAPIWrapper(OneDriveAuth):
 				files or "ChooseNewName" to let OneDrive derive some similar
 				unique name. Behavior of this option mimics underlying API.
 
-			downsize is a true/false API flag, similar to overwrite.'''
+			downsize is a true/false API flag, similar to overwrite.
 
+			bits_api_fallback can be either True/False or an integer (number of
+				bytes), and determines whether method will fall back to using BITS API
+				(as implemented by "put_bits" method) for large files. Default "True"
+				(bool) value will use non-BITS file size limit (api_put_max_bytes, ~100 MiB)
+				as a fallback threshold, passing False will force using single-request uploads.'''
 		overwrite = self._translate_api_flag(overwrite, 'overwrite', ['ChooseNewName'])
 		downsize = self._translate_api_flag(downsize, 'downsize')
 
@@ -279,10 +321,27 @@ class OneDriveAPIWrapper(OneDriveAuth):
 			if isinstance(path_or_tuple, types.StringTypes)\
 			else (path_or_tuple[0], path_or_tuple[1])
 
+		if not isinstance(bits_api_fallback, (int, float, long)):
+			bits_api_fallback = bool(bits_api_fallback)
+		if bits_api_fallback is not False:
+			if bits_api_fallback is True: bits_api_fallback = self.api_put_max_bytes
+			src.seek(0, os.SEEK_END)
+			if src.tell() > bits_api_fallback:
+				return self.put_bits( path_or_tuple,
+					folder_id=folder_id, overwrite=overwrite, downsize=downsize )
+
 		return self( ujoin(folder_id, 'files', name),
 			dict(downsize_photo_uploads=downsize, overwrite=overwrite),
 			data=src, method='put', auth_header=True )
 
+	def put_bits(self, path_or_tuple, folder_id='me/skydrive', overwrite=None, downsize=None):
+		'''Upload a file (object) using BITS API (via several http requests), possibly
+				overwriting (default behavior) a file with the same "name" attribute, if it exists.
+			Not implemented yet.'''
+		overwrite = self._translate_api_flag(overwrite, 'overwrite', ['ChooseNewName'])
+		downsize = self._translate_api_flag(downsize, 'downsize')
+		raise NotImplementedError
+
 	def mkdir(self, name=None, folder_id='me/skydrive', metadata=dict()):
 		'''Create a folder with a specified "name" attribute.
 				folder_id allows to specify a parent folder.
@@ -324,7 +383,7 @@ class OneDriveAPIWrapper(OneDriveAuth):
 		return self.copy(obj_id, folder_id, move=True)
 
 	def comments(self, obj_id):
-		'Get OneDrive object, representing a list of comments for an object.'
+		'Get OneDrive object representing a list of comments for an object.'
 		return self(ujoin(obj_id, 'comments'))
 
 	def comment_add(self, obj_id, message):
